-- zlib.ew -- interface of the 'zlib' general purpose compression library
--  version 1.1.4, March 11th, 2002
--
--  Copyright (C) 1995-2002 Jean-loup Gailly and Mark Adler
--
--  This software is provided 'as-is', without any express or implied
--  warranty.  In no event will the authors be held liable for any damages
--  arising from the use of this software.
--
--  Permission is granted to anyone to use this software for any purpose,
--  including commercial applications, and to alter it and redistribute it
--  freely, subject to the following restrictions:
--
--  1. The origin of this software must not be misrepresented; you must not
--     claim that you wrote the original software. If you use this software
--     in a product, an acknowledgment in the product documentation would be
--     appreciated but is not required.
--  2. Altered source versions must be plainly marked as such, and must not be
--     misrepresented as being the original software.
--  3. This notice may not be removed or altered from any source distribution.
--
--  Jean-loup Gailly        Mark Adler
--  jloup@gzip.org          madler@alumni.caltech.edu
--
--
--  The data format used by the zlib library is described by RFCs (Request for
--  Comments) 1950 to 1952 in the files ftp://ds.internic.net/rfc/rfc1950.txt
--  (zlib format), rfc1951.txt (deflate format) and rfc1952.txt (gzip format).
--

without warning
include machine.e
include dll.e
atom ZLIBDLL
if platform() = 2 then
	ZLIBDLL = open_dll("ZLib.dll")
elsif platform() = 3 then
	ZLIBDLL = open_dll("libz.so")
end if

global constant ZLIB_VERSION = "1.1.4"

--     The 'zlib' compression library provides in-memory compression and
--  decompression functions, including integrity checks of the uncompressed
--  data.  This version of the library supports only one compression method
--  (deflation) but other algorithms will be added later and will have the same
--  stream interface.

--     Compression can be done in a single step if the buffers are large
--  enough (for example if an input file is mmap'ed), or can be done by
--  repeated calls of the compression function.  In the latter case, the
--  application must provide more input and/or consume the output
--  (providing more output space) before each call.

--     The library also supports reading and writing files in gzip (.gz) format
--  with an interface similar to that of stdio.

--     The library does not install any signal handler. The decoder checks
--  the consistency of the compressed data, so the library should never
--  crash even in case of corrupted input.

global type gzFile(object f)
	return atom(f)
end type

--function alloc_func(atom opaque, integer items, integer size)
--function free_func(atom opaque, atom address)

global constant	--struct z_stream_s {
Znext_in = 0,	-- Bytef*-- pointer to next input byte
Zavail_in = 4,	-- uInt  -- number of bytes available at Znext_in
Ztotal_in = 8,	-- uLong -- total number of input bytes read so far
				--
Znext_out = 12,	-- Bytef*-- next output byte should be put here (at the place pointed to)
Zavail_out = 16,-- uInt  -- remaining free space at Znext_out
Ztotal_out = 20,-- uLong -- total number of bytes output so far
				--
Zmsg = 24,		-- char* -- last error message, NULL if no error
Zstate = 28,	-- struct internal_state*-- not visible by applications
				--
Zalloc = 32,	-- alloc_func -- used to allocate the internal state
Zfree = 36,		-- free_func  -- used to free the internal state
Zopaque = 40,	-- voidpf     -- private data object passed to Zalloc and Zfree
				--
Zdata_type = 44,-- int   -- best guess about the data type: ascii or binary
Zadler = 48,	-- uLong -- adler32 value of the uncompressed data
Zreserved = 52	-- uLong -- reserved for future use
				--} z_stream

global function z_stream(object setup)
atom lpstream
lpstream = allocate(14 * 4)
mem_set(lpstream, 0, 14 * 4)
if sequence(setup) then
	if length(setup) = 14 then
		poke4(lpstream, setup)
	end if
end if
	return lpstream
end function

global type z_streamp(object z)
	return atom(z)
end type

--   The application must update Znext_in and Zavail_in when Zavail_in has
--   dropped to zero. It must update Znext_out and Zavail_out when Zavail_out
--   has dropped to zero. The application must initialize Zalloc, Zfree and
--   Zopaque before calling the init function. All other fields are set by the
--   compression library and must not be updated by the application.

--   The Zopaque value provided by the application will be passed as the first
--   parameter for calls of Zalloc and Zfree. This can be useful for custom
--   memory management. The compression library attaches no meaning to the
--   Zopaque value.

--   Zalloc must return Z_NULL if there is not enough memory for the object.
--   If zlib is used in a multi-threaded application, Zalloc and Zfree must be
--   thread safe.

--   The fields Ztotal_in and Ztotal_out can be used for statistics or
--   progress reports. After compression, Ztotal_in holds the total size of
--   the uncompressed data and may be saved for use in the decompressor
--   (particularly if the decompressor wants to decompress everything in
--   a single step).

------------------------------- Constants -------------------------------------

-- Allowed flush values; see deflate() below for details
global constant Z_NO_FLUSH = 0,
				Z_PARTIAL_FLUSH = 1, -- will be removed, use Z_SYNC_FLUSH instead
				Z_SYNC_FLUSH = 2,
				Z_FULL_FLUSH = 3,
				Z_FINISH = 4

global constant SEEK_SET = 0,	-- Seek from beginning of file.
				SEEK_CUR = 1,	-- Seek from current position.
				SEEK_END = 2	-- Set file pointer to EOF plus "offset"

-- * Return codes for the compression/decompression functions. Negative
-- * values are errors, positive values are used for special but normal events.
-- *
global constant Z_OK = 0,
				Z_STREAM_END = 1,
				Z_NEED_DICT = 2,
				Z_ERRNO = -1,
				Z_STREAM_ERROR = -2,
				Z_DATA_ERROR = -3,
				Z_MEM_ERROR = -4,
				Z_BUF_ERROR = -5,
				Z_VERSION_ERROR = -6

-- Compression Levels
global constant Z_NO_COMPRESSION = 0,
				Z_BEST_SPEED = 1,
				Z_BEST_COMPRESSION = 9,
				Z_DEFAULT_COMPRESSION = -1

-- Compression Strategy; see deflateInit2() below for details
global constant Z_FILTERED = 1,
				Z_HUFFMAN_ONLY = 2,
				Z_DEFAULT_STRATEGY = 0

-- Possible values of the Zdata_type field
global constant Z_BINARY = 0,
				Z_ASCII = 1,
				Z_UNKNOWN = 2

-- The deflate compression method (the only one supported in this version)
global constant Z_DEFLATED = 8

global constant Z_NULL = 0  -- for initializing Zalloc, Zfree, Zopaque

-------------------------- basic functions ----------------------------
constant
ZlibVersion = define_c_func(ZLIBDLL, "zlibVersion", {}, C_POINTER),
DeflateInit = define_c_func(ZLIBDLL, "deflateInit_", {C_POINTER, C_INT, C_POINTER, C_INT}, C_INT),
Deflate = define_c_func(ZLIBDLL, "deflate", {C_POINTER, C_INT}, C_INT),
DeflateEnd = define_c_func(ZLIBDLL, "deflateEnd", {C_POINTER}, C_INT),
InflateInit = define_c_func(ZLIBDLL, "inflateInit_", {C_POINTER, C_POINTER, C_INT}, C_INT),
Inflate = define_c_func(ZLIBDLL, "inflate", {C_POINTER, C_INT}, C_INT),
InflateEnd = define_c_func(ZLIBDLL, "inflateEnd", {C_POINTER}, C_INT),
DeflateInit2 = define_c_func(ZLIBDLL, "deflateInit2_", {C_POINTER, C_INT, C_INT, C_INT, C_INT, C_INT, C_POINTER, C_INT}, C_INT),
DeflateSetDictionary = define_c_func(ZLIBDLL, "deflateSetDictionary", {C_POINTER, C_POINTER, C_UINT}, C_INT),
DeflateCopy = define_c_func(ZLIBDLL, "deflateCopy", {C_POINTER, C_POINTER}, C_INT),
DeflateReset = define_c_func(ZLIBDLL, "deflateReset", {C_POINTER}, C_INT),
DeflateParams = define_c_func(ZLIBDLL, "deflateParams", {C_POINTER, C_INT, C_INT}, C_INT),
InflateInit2 = define_c_func(ZLIBDLL, "inflateInit2_", {C_POINTER, C_INT, C_POINTER, C_INT}, C_INT),
InflateSetDictionary = define_c_func(ZLIBDLL, "inflateSetDictionary", {C_POINTER, C_POINTER, C_UINT}, C_INT),
InflateSync = define_c_func(ZLIBDLL, "inflateSync", {C_POINTER}, C_INT),
InflateReset = define_c_func(ZLIBDLL, "inflateReset", {C_POINTER}, C_INT),
Compress = define_c_func(ZLIBDLL, "compress", {C_POINTER, C_POINTER, C_POINTER, C_ULONG}, C_INT),
Compress2 = define_c_func(ZLIBDLL, "compress2", {C_POINTER, C_POINTER, C_POINTER, C_ULONG, C_INT}, C_INT),
Uncompress = define_c_func(ZLIBDLL, "uncompress", {C_POINTER, C_POINTER, C_POINTER, C_ULONG}, C_INT),
Gzopen = define_c_func(ZLIBDLL, "gzopen", {C_POINTER, C_POINTER}, C_POINTER),
Gzdopen = define_c_func(ZLIBDLL, "gzdopen", {C_INT, C_POINTER}, C_POINTER),
Gzsetparams = define_c_func(ZLIBDLL, "gzsetparams", {C_POINTER, C_INT, C_INT}, C_INT),
Gzread = define_c_func(ZLIBDLL, "gzread", {C_POINTER, C_POINTER, C_UINT}, C_INT),
Gzwrite = define_c_func(ZLIBDLL, "gzwrite", {C_POINTER, C_POINTER, C_UINT}, C_INT),
--Gzprintf = define_c_func(ZLIBDLL, "+_gzprintf", {C_POINTER, C_POINTER, ...}, C_INT),
Gzputs = define_c_func(ZLIBDLL, "gzputs", {C_POINTER, C_POINTER}, C_INT),
Gzgets = define_c_func(ZLIBDLL, "gzgets", {C_POINTER, C_POINTER, C_INT}, C_POINTER),
Gzputc = define_c_func(ZLIBDLL, "gzputc", {C_POINTER, C_INT}, C_INT),
Gzgetc = define_c_func(ZLIBDLL, "gzgetc", {C_POINTER}, C_INT),
Gzflush = define_c_func(ZLIBDLL, "gzflush", {C_POINTER, C_INT}, C_INT),
Gzseek = define_c_func(ZLIBDLL, "gzseek", {C_POINTER, C_LONG, C_INT}, C_LONG),
Gzrewind = define_c_func(ZLIBDLL, "gzrewind", {C_POINTER}, C_INT),
Gztell = define_c_func(ZLIBDLL, "gztell", {C_POINTER}, C_LONG),
Gzeof = define_c_func(ZLIBDLL, "gzeof", {C_POINTER}, C_INT),
Gzclose = define_c_func(ZLIBDLL, "gzclose", {C_POINTER}, C_INT),
Gzerror = define_c_func(ZLIBDLL, "gzerror", {C_POINTER, C_POINTER}, C_POINTER),
Adler32 = define_c_func(ZLIBDLL, "adler32", {C_ULONG, C_POINTER, C_UINT}, C_ULONG),
Crc32 = define_c_func(ZLIBDLL, "crc32", {C_ULONG, C_POINTER, C_UINT}, C_ULONG),
ZError = define_c_func(ZLIBDLL, "zError", {C_INT}, C_POINTER),
InflateSyncPoint = define_c_func(ZLIBDLL, "inflateSyncPoint", {C_POINTER}, C_INT),
Get_crc_table = define_c_func(ZLIBDLL, "get_crc_table", {}, C_POINTER)

function alloc8(atom len)
atom ptr
ptr = allocate(len)
mem_set(ptr, 0, len)
	return ptr
end function

----- zlibVersion -----
-- The application can compare zlibVersion and ZLIB_VERSION for consistency.
-- If the first character differs, the library code actually used is
-- not compatible with the zlib.e header file used by the application.
-- This check is automatically made by deflateInit and inflateInit.
global function zlibVersion()
atom ret
integer len
len = 0
ret = c_func(ZlibVersion, {})
while peek(ret + len) do
	len += 1
end while
	return peek({ret, len})
end function

----- deflateInit -----
--   Initializes the internal stream Zstate for compression. The fields
-- Zalloc, Zfree and Zopaque must be initialized before by the caller.
-- If Zalloc and Zfree are set to Z_NULL, deflateInit updates them to
-- use default allocation functions.
--
--   The compression level must be Z_DEFAULT_COMPRESSION, or between 0 and 9:
-- 1 gives best speed, 9 gives best compression, 0 gives no compression at
-- all (the input data is simply copied a block at a time).
-- Z_DEFAULT_COMPRESSION requests a default compromise between speed and
-- compression (currently equivalent to level 6).
--
--   deflateInit returns Z_OK if success, Z_MEM_ERROR if there was not
-- enough memory, Z_STREAM_ERROR if level is not a valid compression level,
-- Z_VERSION_ERROR if the zlib library version (zlib_version) is incompatible
-- with the version assumed by the caller (ZLIB_VERSION).
-- Zmsg is set to null if there is no error message.  deflateInit does not
-- perform any compression: this will be done by deflate().
global function deflateInit_(z_streamp strm, integer level, sequence version, integer stream_size)
atom lpvers
integer ret
lpvers = allocate_string(version)
ret = c_func(DeflateInit, {strm, level, lpvers, stream_size})
free(lpvers)
	return ret
end function

global function deflateInit(z_streamp strm, integer level)
	return deflateInit_(strm, level, ZLIB_VERSION, 14 * 4)
end function

----- deflate -----
--   deflate compresses as much data as possible, and stops when the input
-- buffer becomes empty or the output buffer becomes full. It may introduce some
-- output latency (reading input without producing any output) except when
-- forced to flush.
--
--   The detailed semantics are as follows. deflate performs one or both of the
-- following actions:
--
-- * Compress more input starting at Znext_in and update Znext_in and Zavail_in
--   accordingly. If not all input can be processed (because there is not
--   enough room in the output buffer), Znext_in and Zavail_in are updated and
--   processing will resume at this point for the next call of deflate().
--
-- * Provide more output starting at Znext_out and update Znext_out and Zavail_out
--   accordingly. This action is forced if the parameter flush is non zero.
--   Forcing flush frequently degrades the compression ratio, so this parameter
--   should be set only when necessary (in interactive applications).
--   Some output may be provided even if flush is not set.
--
-- Before the call of deflate(), the application should ensure that at least
-- one of the actions is possible, by providing more input and/or consuming
-- more output, and updating Zavail_in or Zavail_out accordingly; Zavail_out
-- should never be zero before the call. The application can consume the
-- compressed output when it wants, for example when the output buffer is full
-- (Zavail_out = 0), or after each call of deflate(). If deflate returns Z_OK
-- and with zero Zavail_out, it must be called again after making room in the
-- output buffer because there might be more output pending.
--
--   If the parameter flush is set to Z_SYNC_FLUSH, all pending output is
-- flushed to the output buffer and the output is aligned on a byte boundary, so
-- that the decompressor can get all input data available so far. (In particular
-- Zavail_in is zero after the call if enough output space has been provided
-- before the call.)  Flushing may degrade compression for some compression
-- algorithms and so it should be used only when necessary.
--
--   If flush is set to Z_FULL_FLUSH, all output is flushed as with
-- Z_SYNC_FLUSH, and the compression Zstate is reset so that decompression can
-- restart from this point if previous compressed data has been damaged or if
-- random access is desired. Using Z_FULL_FLUSH too often can seriously degrade
-- the compression.
--
--   If deflate returns with Zavail_out == 0, this function must be called again
-- with the same value of the flush parameter and more output space (updated
-- Zavail_out), until the flush is complete (deflate returns with non-zero
-- Zavail_out).
--
--   If the parameter flush is set to Z_FINISH, pending input is processed,
-- pending output is flushed and deflate returns with Z_STREAM_END if there
-- was enough output space; if deflate returns with Z_OK, this function must be
-- called again with Z_FINISH and more output space (updated Zavail_out) but no
-- more input data, until it returns with Z_STREAM_END or an error. After
-- deflate has returned Z_STREAM_END, the only possible operations on the
-- stream are deflateReset or deflateEnd.
--
--   Z_FINISH can be used immediately after deflateInit if all the compression
-- is to be done in a single step. In this case, Zavail_out must be at least
-- 0.1% larger than Zavail_in plus 12 bytes.  If deflate does not return
-- Z_STREAM_END, then it must be called again as described above.
--
--   deflate() sets strm->Zadler to the adler32 checksum of all input read
-- so far (that is, Ztotal_in bytes).
--
--   deflate() may update Zdata_type if it can make a good guess about
-- the input data type (Z_ASCII or Z_BINARY). In doubt, the data is considered
-- binary. This field is only for information purposes and does not affect
-- the compression algorithm in any manner.
--
--   deflate() returns Z_OK if some progress has been made (more input
-- processed or more output produced), Z_STREAM_END if all input has been
-- consumed and all output has been produced (only when flush is set to
-- Z_FINISH), Z_STREAM_ERROR if the stream state was inconsistent (for example
-- if Znext_in or Znext_out was NULL), Z_BUF_ERROR if no progress is possible
-- (for example Zavail_in or Zavail_out was zero).
global function deflate(z_streamp strm, integer flush)
	return c_func(Deflate, {strm, flush})
end function

----- deflateEnd -----
--   All dynamically allocated data structures for this stream are freed.
-- This function discards any unprocessed input and does not flush any
-- pending output.
--
--   deflateEnd returns Z_OK if success, Z_STREAM_ERROR if the
-- stream Zstate was inconsistent, Z_DATA_ERROR if the stream was freed
-- prematurely (some input or output was discarded). In the error case,
-- Zmsg may be set but then points to a static string (which must not be
-- deallocated).
global function deflateEnd(z_streamp strm)
	return c_func(DeflateEnd, {strm})
end function

----- inflateInit -----
--   Initializes the internal stream Zstate for decompression. The fields
-- Znext_in, Zavail_in, Zalloc, Zfree and Zopaque must be initialized before by
-- the caller. If Znext_in is not Z_NULL and Zavail_in is large enough (the exact
-- value depends on the compression method), inflateInit determines the
-- compression method from the zlib header and allocates all data structures
-- accordingly; otherwise the allocation will be deferred to the first call of
-- inflate.  If Zalloc and Zfree are set to Z_NULL, inflateInit updates them to
-- use default allocation functions.
--
--   inflateInit returns Z_OK if success, Z_MEM_ERROR if there was not enough
-- memory, Z_VERSION_ERROR if the zlib library version is incompatible with the
-- version assumed by the caller.  Zmsg is set to null if there is no error
-- message. inflateInit does not perform any decompression apart from reading
-- the zlib header if present: this will be done by inflate().  (So Znext_in and
-- Zavail_in may be modified, but Znext_out and Zavail_out are unchanged.)
global function inflateInit_(z_streamp strm, sequence version, integer stream_size)
atom lpvers
integer ret
lpvers = allocate_string(version)
ret = c_func(InflateInit, {strm, lpvers, stream_size})
free(lpvers)
	return ret
end function

global function inflateInit(z_streamp strm)
	return inflateInit_(strm, ZLIB_VERSION, 14 * 4)
end function

----- inflate -----
--   inflate decompresses as much data as possible, and stops when the input
-- buffer becomes empty or the output buffer becomes full. It may some
-- introduce some output latency (reading input without producing any output)
-- except when forced to flush.
--
-- The detailed semantics are as follows. inflate performs one or both of the
-- following actions:
--
-- * Decompress more input starting at Znext_in and update Znext_in and Zavail_in
--   accordingly. If not all input can be processed (because there is not
--   enough room in the output buffer), Znext_in is updated and processing
--   will resume at this point for the next call of inflate().
--
-- * Provide more output starting at Znext_out and update Znext_out and Zavail_out
--   accordingly.  inflate() provides as much output as possible, until there
--   is no more input data or no more space in the output buffer (see below
--   about the flush parameter).
--
-- Before the call of inflate(), the application should ensure that at least
-- one of the actions is possible, by providing more input and/or consuming
-- more output, and updating the Znext_* and Zavail_* values accordingly.
-- The application can consume the uncompressed output when it wants, for
-- example when the output buffer is full (Zavail_out == 0), or after each
-- call of inflate(). If inflate returns Z_OK and with zero Zavail_out, it
-- must be called again after making room in the output buffer because there
-- might be more output pending.
--
--   If the parameter flush is set to Z_SYNC_FLUSH, inflate flushes as much
-- output as possible to the output buffer. The flushing behavior of inflate is
-- not specified for values of the flush parameter other than Z_SYNC_FLUSH
-- and Z_FINISH, but the current implementation actually flushes as much output
-- as possible anyway.
--
--   inflate() should normally be called until it returns Z_STREAM_END or an
-- error. However if all decompression is to be performed in a single step
-- (a single call of inflate), the parameter flush should be set to
-- Z_FINISH. In this case all pending input is processed and all pending
-- output is flushed; Zavail_out must be large enough to hold all the
-- uncompressed data. (The size of the uncompressed data may have been saved
-- by the compressor for this purpose.) The next operation on this stream must
-- be inflateEnd to deallocate the decompression Zstate. The use of Z_FINISH
-- is never required, but can be used to inform inflate that a faster routine
-- may be used for the single inflate() call.
--
--    If a preset dictionary is needed at this point (see inflateSetDictionary
-- below), inflate sets strm->Zadler to the adler32 checksum of the
-- dictionary chosen by the compressor and returns Z_NEED_DICT; otherwise 
-- it sets strm->Zadler to the adler32 checksum of all output produced
-- so far (that is, Ztotal_out bytes) and returns Z_OK, Z_STREAM_END or
-- an error code as described below. At the end of the stream, inflate()
-- checks that its computed adler32 checksum is equal to that saved by the
-- compressor and returns Z_STREAM_END only if the checksum is correct.
--
--   inflate() returns Z_OK if some progress has been made (more input processed
-- or more output produced), Z_STREAM_END if the end of the compressed data has
-- been reached and all uncompressed output has been produced, Z_NEED_DICT if a
-- preset dictionary is needed at this point, Z_DATA_ERROR if the input data was
-- corrupted (input stream not conforming to the zlib format or incorrect
-- adler32 checksum), Z_STREAM_ERROR if the stream structure was inconsistent
-- (for example if Znext_in or Znext_out was NULL), Z_MEM_ERROR if there was not
-- enough memory, Z_BUF_ERROR if no progress is possible or if there was not
-- enough room in the output buffer when Z_FINISH is used. In the Z_DATA_ERROR
-- case, the application may then call inflateSync to look for a good
-- compression block.
global function inflate(z_streamp strm, integer flush)
	return c_func(Inflate, {strm, flush})
end function

----- inflateEnd -----
--   All dynamically allocated data structures for this stream are freed.
-- This function discards any unprocessed input and does not flush any
-- pending output.
--
--   inflateEnd returns Z_OK if success, Z_STREAM_ERROR if the stream Zstate
-- was inconsistent. In the error case, Zmsg may be set but then points to a
-- static string (which must not be deallocated).
global function inflateEnd(z_streamp strm)
	return c_func(InflateEnd, {strm})
end function

-------------------------- Advanced Functions ---------------------------------
--                                                                           --
--   The following functions are needed only in some special applications.   --
--                                                                           --
-------------------------------------------------------------------------------

----- deflateInit2 -----
--   This is another version of deflateInit with more compression options. The
-- fields Znext_in, Zalloc, Zfree and Zopaque must be initialized before by
-- the caller.
--
--   The method parameter is the compression method. It must be Z_DEFLATED in
-- this version of the library.
--
--   The windowBits parameter is the base two logarithm of the window size
-- (the size of the history buffer).  It should be in the range 8..15 for this
-- version of the library. Larger values of this parameter result in better
-- compression at the expense of memory usage. The default value is 15 if
-- deflateInit is used instead.
--
--   The memLevel parameter specifies how much memory should be allocated
-- for the internal compression Zstate. memLevel=1 uses minimum memory but
-- is slow and reduces compression ratio; memLevel=9 uses maximum memory
-- for optimal speed. The default value is 8.
--
--   The strategy parameter is used to tune the compression algorithm. Use the
-- value Z_DEFAULT_STRATEGY for normal data, Z_FILTERED for data produced by a
-- filter (or predictor), or Z_HUFFMAN_ONLY to force Huffman encoding only (no
-- string match).  Filtered data consists mostly of small values with a
-- somewhat random distribution. In this case, the compression algorithm is
-- tuned to compress them better. The effect of Z_FILTERED is to force more
-- Huffman coding and less string matching; it is somewhat intermediate
-- between Z_DEFAULT and Z_HUFFMAN_ONLY. The strategy parameter only affects
-- the compression ratio but not the correctness of the compressed output even
-- if it is not set appropriately.
--
--    deflateInit2 returns Z_OK if success, Z_MEM_ERROR if there was not enough
-- memory, Z_STREAM_ERROR if a parameter is invalid (such as an invalid
-- method). Zmsg is set to null if there is no error message.  deflateInit2 does
-- not perform any compression: this will be done by deflate().
global function deflateInit2_(z_streamp strm, integer level, integer method, integer windowBits, integer memLevel, integer strategy, sequence version, integer stream_size)
atom lpvers
integer ret
lpvers = allocate_string(version)
ret = c_func(DeflateInit2, {strm, level, method, windowBits, memLevel, strategy, lpvers, stream_size})
free(lpvers)
	return ret
end function

global function deflateInit2(z_streamp strm, integer level, integer method, integer windowBits, integer memLevel, integer strategy)
	return deflateInit2_(strm, level, method, windowBits, memLevel, strategy, ZLIB_VERSION, 14 * 4)
end function

----- deflateSetDictionary -----
--   Initializes the compression dictionary from the given byte sequence
-- without producing any compressed output. This function must be called
-- immediately after deflateInit, deflateInit2 or deflateReset, before any
-- call of deflate. The compressor and decompressor must use exactly the same
-- dictionary (see inflateSetDictionary).
--
--   The dictionary should consist of strings (byte sequences) that are likely
-- to be encountered later in the data to be compressed, with the most commonly
-- used strings preferably put towards the end of the dictionary. Using a
-- dictionary is most useful when the data to be compressed is short and can be
-- predicted with good accuracy; the data can then be compressed better than
-- with the default empty dictionary.
--
--   Depending on the size of the compression data structures selected by
-- deflateInit or deflateInit2, a part of the dictionary may in effect be
-- discarded, for example if the dictionary is larger than the window size in
-- deflate or deflate2. Thus the strings most likely to be useful should be
-- put at the end of the dictionary, not at the front.
--
--   Upon return of this function, strm->Zadler is set to the Adler32 value
-- of the dictionary; the decompressor may later use this value to determine
-- which dictionary has been used by the compressor. (The Adler32 value
-- applies to the whole dictionary even if only a subset of the dictionary is
-- actually used by the compressor.)
--
--   deflateSetDictionary returns Z_OK if success, or Z_STREAM_ERROR if a
-- parameter is invalid (such as NULL dictionary) or the stream Zstate is
-- inconsistent (for example if deflate has already been called for this stream
-- or if the compression method is bsort). deflateSetDictionary does not
-- perform any compression: this will be done by deflate().
global function deflateSetDictionary(z_streamp strm, sequence dictionary)
integer dictLength, ret
atom lpDict
dictLength = length(dictionary)
lpDict = alloc8(dictLength)
poke(lpDict, dictionary)
ret = c_func(DeflateSetDictionary, {strm, lpDict, dictLength})
free(lpDict)
	return ret
end function

----- deflateCopy -----
--   Sets the destination stream as a complete copy of the source stream.
--
--   This function can be useful when several compression strategies will be
-- tried, for example when there are several ways of pre-processing the input
-- data with a filter. The streams that will be discarded should then be freed
-- by calling deflateEnd.  Note that deflateCopy duplicates the internal
-- compression Zstate which can be quite large, so this strategy is slow and
-- can consume lots of memory.
--
--   deflateCopy returns Z_OK if success, Z_MEM_ERROR if there was not
-- enough memory, Z_STREAM_ERROR if the source stream state was inconsistent
-- (such as Zalloc being NULL). Zmsg is left unchanged in both source and
-- destination.
global function deflateCopy(z_streamp dest, z_streamp source)
	return c_func(DeflateCopy, {dest, source})
end function

----- deflateReset -----
--   This function is equivalent to deflateEnd followed by deflateInit,
-- but does not free and reallocate all the internal compression Zstate.
-- The stream will keep the same compression level and any other attributes
-- that may have been set by deflateInit2.
--
--    deflateReset returns Z_OK if success, or Z_STREAM_ERROR if the source
-- stream Zstate was inconsistent (such as Zalloc or state being NULL).
global function deflateReset(z_streamp strm)
	return c_func(DeflateReset, {strm})
end function

----- deflateParams -----
--   Dynamically update the compression level and compression strategy.  The
-- interpretation of level and strategy is as in deflateInit2.  This can be
-- used to switch between compression and straight copy of the input data, or
-- to switch to a different kind of input data requiring a different
-- strategy. If the compression level is changed, the input available so far
-- is compressed with the old level (and may be flushed); the new level will
-- take effect only at the next call of deflate().
--
--   Before the call of deflateParams, the stream state must be set as for
-- a call of deflate(), since the currently available input may have to
-- be compressed and flushed. In particular, strm->Zavail_out must be non-zero.
--
--   deflateParams returns Z_OK if success, Z_STREAM_ERROR if the source
-- stream state was inconsistent or if a parameter was invalid, Z_BUF_ERROR
-- if strm->Zavail_out was zero.
global function deflateParams(z_streamp strm, integer level, integer strategy)
	return c_func(DeflateParams, {strm, level, strategy})
end function

----- inflateInit2 -----
--   This is another version of inflateInit with an extra parameter. The
-- fields Znext_in, Zavail_in, Zalloc, Zfree and Zopaque must be initialized
-- before by the caller.
--
--   The windowBits parameter is the base two logarithm of the maximum window
-- size (the size of the history buffer).  It should be in the range 8..15 for
-- this version of the library. The default value is 15 if inflateInit is used
-- instead. If a compressed stream with a larger window size is given as
-- input, inflate() will return with the error code Z_DATA_ERROR instead of
-- trying to allocate a larger window.
--
--    inflateInit2 returns Z_OK if success, Z_MEM_ERROR if there was not enough
-- memory, Z_STREAM_ERROR if a parameter is invalid (such as a negative
-- memLevel). Zmsg is set to null if there is no error message.  inflateInit2
-- does not perform any decompression apart from reading the zlib header if
-- present: this will be done by inflate(). (So Znext_in and Zavail_in may be
-- modified, but Znext_out and Zavail_out are unchanged.)
global function inflateInit2_(z_streamp strm, integer windowBits, sequence version, integer stream_size)
atom lpvers
integer ret
lpvers = allocate_string(version)
ret = c_func(InflateInit2, {strm, windowBits, lpvers, stream_size})
free(lpvers)
	return ret
end function

global function inflateInit2(z_streamp strm, integer windowBits)
	return inflateInit2_(strm, windowBits, ZLIB_VERSION, 14 * 4)
end function

----- inflateSetDictionary -----
--   Initializes the decompression dictionary from the given uncompressed byte
-- sequence. This function must be called immediately after a call of inflate
-- if this call returned Z_NEED_DICT. The dictionary chosen by the compressor
-- can be determined from the Adler32 value returned by this call of
-- inflate. The compressor and decompressor must use exactly the same
-- dictionary (see deflateSetDictionary).
--
--   inflateSetDictionary returns Z_OK if success, Z_STREAM_ERROR if a
-- parameter is invalid (such as NULL dictionary) or the stream Zstate is
-- inconsistent, Z_DATA_ERROR if the given dictionary doesn't match the
-- expected one (incorrect Adler32 value). inflateSetDictionary does not
-- perform any decompression: this will be done by subsequent calls of
-- inflate().
global function inflateSetDictionary(z_streamp strm, sequence dictionary)
integer dictLength, ret
atom lpDict
dictLength = length(dictionary)
lpDict = alloc8(dictLength)
poke(lpDict, dictionary)
ret = c_func(InflateSetDictionary, {strm, lpDict, dictLength})
free(lpDict)
	return ret
end function

----- inflateSync -----
--   Skips invalid compressed data until a full flush point (see above the
-- description of deflate with Z_FULL_FLUSH) can be found, or until all
-- available input is skipped. No output is provided.
--
--   inflateSync returns Z_OK if a full flush point has been found, Z_BUF_ERROR
-- if no more input was provided, Z_DATA_ERROR if no flush point has been found,
-- or Z_STREAM_ERROR if the stream structure was inconsistent. In the success
-- case, the application may save the current current value of Ztotal_in which
-- indicates where valid compressed data was found. In the error case, the
-- application may repeatedly call inflateSync, providing more input each time,
-- until success or end of the input data.
global function inflateSync(z_streamp strm)
	return c_func(InflateSync, {strm})
end function

----- inflateReset -----
--   This function is equivalent to inflateEnd followed by inflateInit,
-- but does not free and reallocate all the internal decompression state.
-- The stream will keep attributes that may have been set by inflateInit2.
--
--    inflateReset returns Z_OK if success, or Z_STREAM_ERROR if the source
-- stream state was inconsistent (such as Zalloc or state being NULL).
global function inflateReset(z_streamp strm)
	return c_func(InflateReset, {strm})
end function

-------------------------- Utility Functions ----------------------------------
--                                                                           --
--                                                                           --
--     The following utility functions are implemented on top of the         --
--   basic stream-oriented functions. To simplify the interface, some        --
--   default options are assumed (compression level and memory usage,        --
--   standard memory allocation functions). The source code of these         --
--   utility functions can easily be modified if you need special options.   --
--                                                                           --
-------------------------------------------------------------------------------

----- compress -----
--   Compresses the source buffer into the destination buffer. Returns {len, buffer},
-- where len is the length of buffer. If buffer is equal to "", then check len for
-- the error code (if there is one).
--
--   This function can be used to compress a whole file at once if the
-- input file is entirely read into the source sequence.
--
--   compress returns Z_OK if success, Z_MEM_ERROR if there was not
-- enough memory, Z_BUF_ERROR if there was not enough room in the output
-- buffer.
global function compress(sequence source)
atom lpdest, lpsrc, lpdestlen
integer destLen, sourceLen, ret
sequence dest
sourceLen = length(source)
destLen = floor((sourceLen * 1.001) + 12 + 0.9)
lpsrc = alloc8(sourceLen)
poke(lpsrc, source)
lpdestlen = alloc8(4)
poke4(lpdestlen, destLen)
lpdest = alloc8(destLen)
ret = c_func(Compress, {lpdest, lpdestlen, lpsrc, sourceLen})
if ret = Z_OK then
	destLen = peek4u(lpdestlen)
	dest = peek({lpdest, destLen})
else
	destLen = ret
	dest = {}
end if
free(lpsrc)
free(lpdestlen)
free(lpdest)
	return {destLen, dest}
end function

----- compress2 -----
--   Compresses the source buffer into the destination buffer. The level
-- parameter has the same meaning as in deflateInit. Returns {len, buffer},
-- where len is the length of buffer. If buffer is equal to "", then check len for
-- the error code (if there is one).
--
--   compress2 returns Z_OK if success, Z_MEM_ERROR if there was not enough
-- memory, Z_BUF_ERROR if there was not enough room in the output buffer,
-- Z_STREAM_ERROR if the level parameter is invalid.
global function compress2(sequence source, integer level)
atom lpdest, lpsrc, lpdestlen
integer destLen, sourceLen, ret
sequence dest
sourceLen = length(source)
destLen = floor((sourceLen * 1.001) + 12 + 0.9)
lpsrc = alloc8(sourceLen)
poke(lpsrc, source)
lpdestlen = alloc8(4)
poke4(lpdestlen, destLen)
lpdest = alloc8(destLen)
ret = c_func(Compress2, {lpdest, lpdestlen, lpsrc, sourceLen, level})
if ret = Z_OK then
	destLen = peek4u(lpdestlen)
	dest = peek({dest, destLen})
else
	destLen = ret
	dest = {}
end if
free(lpsrc)
free(lpdestlen)
free(lpdest)
	return {destLen, dest}
end function

----- uncompress -----
--   Decompresses the source buffer into the destination buffer. Upon entry,
-- destLen is the total size of the destination buffer, which must be large
-- enough to hold the entire uncompressed data. (The size of the uncompressed
-- data must have been saved previously by the compressor and transmitted
-- to the decompressor by some mechanism outside the scope of this compression
-- library.) Returns {len, buffer}, where len is the length of buffer. If buffer
-- is equal to "", then check len for the error code (if there is one).
--
--   This function can be used to decompress a whole file at once if the
-- input file is read into the source sequence.
--
--   uncompress returns Z_OK if success, Z_MEM_ERROR if there was not
-- enough memory, Z_BUF_ERROR if there was not enough room in the output
-- buffer, or Z_DATA_ERROR if the input data was corrupted.
global function uncompress(sequence source, integer destLen)
atom lpdest, lpsrc, lpdestlen
integer sourceLen, ret
sequence dest
sourceLen = length(source)
lpsrc = alloc8(sourceLen)
poke(lpsrc, source)
lpdestlen = alloc8(4)
poke4(lpdestlen, destLen)
lpdest = alloc8(destLen)
ret = c_func(Uncompress, {lpdest, lpdestlen, lpsrc, sourceLen})
if ret = Z_OK then
	destLen = peek4u(lpdestlen)
	dest = peek({lpdest, destLen})
else
	destLen = ret
	dest = {}
end if
free(lpsrc)
free(lpdestlen)
free(lpdest)
	return {destLen, dest}
end function

----- gzopen -----
--   Opens a gzip (.gz) file for reading or writing. The mode parameter
-- is as in open ("rb" or "wb") but can also include a compression level
-- ("wb9") or a strategy: 'f' for filtered data as in "wb6f", 'h' for
-- Huffman only compression as in "wb1h". (See the description
-- of deflateInit2 for more information about the strategy parameter.)
--
--   gzopen can be used to read a file which is not in gzip format; in this
-- case gzread will directly read from the file without decompression.
--
--   gzopen returns NULL if the file could not be opened or if there was
-- insufficient memory to allocate the (de)compression Zstate; errno
-- can be checked to distinguish the two cases (if errno is zero, the
-- zlib error is Z_MEM_ERROR).
global function gzopen(sequence path, sequence mode)
atom lppath, lpmode
gzFile ret
lppath = allocate_string(path)
lpmode = allocate_string(mode)
ret = c_func(Gzopen, {lppath, lpmode})
free(lppath)
free(lpmode)
	return ret
end function

----- gzdopen -----
--   gzdopen() associates a gzFile with the file descriptor fd.  File
-- descriptors are obtained from calls like open, dup, creat, pipe or
-- fileno (in the file has been previously opened with fopen).
-- The mode parameter is as in gzopen.
--   The next call of gzclose on the returned gzFile will also close the
-- file descriptor fd, just like fclose(fdopen(fd), mode) closes the file
-- descriptor fd. If you want to keep fd open, use gzdopen(dup(fd), mode).
--   gzdopen returns NULL if there was insufficient memory to allocate
-- the (de)compression Zstate.
--   NOTE: This function is here for completeness' sake. It will not work with
-- Euphoria as Euphoria does not use file descriptors. However, if you happen
-- to use file descriptors in some other way (not using Euphoria's default
-- open() and close()), then you can still use this function.
global function gzdopen(integer fd, sequence mode)
atom lpmode
gzFile ret
lpmode = allocate_string(mode)
ret = c_func(Gzdopen, {fd, lpmode})
free(lpmode)
	return ret
end function

----- gzsetparams -----
--   Dynamically update the compression level or strategy. See the description
-- of deflateInit2 for the meaning of these parameters.
--   gzsetparams returns Z_OK if success, or Z_STREAM_ERROR if the file was not
-- opened for writing.
global function gzsetparams(gzFile file, integer level, integer strategy)
	return c_func(Gzsetparams, {file, level, strategy})
end function

----- gzread -----
--   Reads the given number of uncompressed bytes from the compressed file.
-- If the input file was not in gzip format, gzread copies the given number
-- of bytes into the buffer.
--   gzread returns the number of uncompressed bytes actually read (in
-- {bytes, buffer}; (0 for end of file, -1 for error))
global function gzread(gzFile file, integer len)
atom lpbuf
sequence buf
integer ret
lpbuf = alloc8(len)
ret = c_func(Gzread, {file, lpbuf, len})
if ret > 0 then
	buf = peek({lpbuf, ret})
else
	buf = {}
end if
free(lpbuf)
	return {ret, buf}
end function

----- gzwrite -----
--   Writes the given number of uncompressed bytes into the compressed file.
-- gzwrite returns the number of uncompressed bytes actually written
-- (0 in case of error).
global function gzwrite(gzFile file, sequence buf)
atom lpbuf
integer ret, len
len = length(buf)
lpbuf = alloc8(len)
poke(lpbuf, buf)
ret = c_func(Gzwrite, {file, lpbuf, len})
free(lpbuf)
	return ret
end function

----- gzprintf -----
--   Converts, formats, and writes the args to the compressed file under
-- control of the format string, as in fprintf. gzprintf returns the number of
-- uncompressed bytes actually written (0 in case of error).
global function gzprintf(gzFile file, sequence format, sequence params)
integer infrmt, ret, xgzprintf
sequence param_form, ptrs
infrmt = 0
param_form = {C_POINTER, C_POINTER}
for loop = 1 to length(format) do
	if format[loop] = '%' then
		infrmt = 1
	elsif (infrmt) then
		ret = format[loop]
		if ((ret >= '0') and (ret <= '9')) or (ret = '.') then
			infrmt = 1
		elsif (ret = '-') or (ret = '+') or (ret = ' ') or (ret = '#') then
			infrmt = 1
		elsif (ret = 'h') or (ret = 'l') then
			infrmt = 1
		elsif (ret = 'd') or (ret = 'i') or (ret = '*') then
			if format[loop - 1] = 'h' then
				param_form &= C_SHORT
			elsif format[loop - 1] = 'l' then
				param_form &= C_LONG
			else
				param_form &= C_INT
			end if
			infrmt = 0
		elsif (ret = 'o') or (ret = 'u') or (ret = 'x') or (ret = 'X') then
			if format[loop - 1] = 'h' then
				param_form &= C_USHORT
			elsif format[loop - 1] = 'l' then
				param_form &= C_ULONG
			else
				param_form &= C_UINT
			end if
			infrmt = 0
		elsif (ret = 'e') or (ret = 'g') or (ret = 'E') or (ret = 'G') then
			if format[loop - 1] = 'l' then
				param_form &= C_DOUBLE
			else
				param_form &= C_FLOAT
			end if
			infrmt = 0
		elsif ret = 'c' then
			param_form &= C_UCHAR
			infrmt = 0
		elsif ret = 's' then
			param_form &= C_POINTER
			infrmt = 0
		elsif (ret = 'p') or (ret = 'n') then
			param_form &= C_POINTER
		elsif ret = '%' then
			infrmt = 0
		else
			infrmt = 0
		end if
	end if
end for
ptrs = {allocate_string(format)}
for loop = 1 to length(params) do
	if sequence(params[loop]) then
		params[loop] = allocate_string(params[loop])
		ptrs &= params[loop]
	end if
end for
if platform() = 2 then
	xgzprintf = define_c_func(ZLIBDLL, "+_gzprintf", param_form, C_INT)
elsif platform() = 3 then
	xgzprintf = define_c_func(ZLIBDLL, "gzprintf", param_form, C_INT)
end if
ret = c_func(xgzprintf, {file, ptrs[1]} & params)
for loop = 1 to length(ptrs) do
	free(ptrs[loop])
end for
	return ret
end function

----- gzputs -----
--    Writes the given null-terminated string to the compressed file, excluding
-- the terminating null character.
--    gzputs returns the number of characters written, or -1 in case of error.
global function gzputs(gzFile file, sequence s)
atom lpseq
integer ret
lpseq = allocate_string(s)
ret = c_func(Gzputs, {file, lpseq})
free(lpseq)
	return ret
end function

----- gzgets -----
--    Reads bytes from the compressed file until len-1 characters are read, or
-- a newline character is read and transferred to buf, or an end-of-file
-- condition is encountered.
--    gzgets returns buf, or "" in case of error.
global function gzgets(gzFile file, integer len)
atom ret, lpbuf
sequence buf
lpbuf = alloc8(len + 1)
ret = c_func(Gzgets, {file, lpbuf, len})
if ret then
	len = 0
	while peek(lpbuf + len) do
		len += 1
	end while
	buf = peek({lpbuf, len})
else
	buf = ""
end if
free(lpbuf)
	return buf
end function

----- gzputc -----
--    Writes c, converted to an unsigned char, into the compressed file.
-- gzputc returns the value that was written, or -1 in case of error.
global function gzputc(gzFile file, integer c)
	return c_func(Gzputc, {file, c})
end function

----- gzgetc -----
--    Reads one byte from the compressed file. gzgetc returns this byte
-- or -1 in case of end of file or error.
global function gzgetc(gzFile file)
	return c_func(Gzgetc, {file})
end function

----- gzflush -----
--   Flushes all pending output into the compressed file. The parameter
-- flush is as in the deflate() function. The return value is the zlib
-- error number (see function gzerror below). gzflush returns Z_OK if
-- the flush parameter is Z_FINISH and all output could be flushed.
--   gzflush should be called only when strictly necessary because it can
-- degrade compression.
global function gzflush(gzFile file, integer flush)
	return c_func(Gzflush, {file, flush})
end function

----- gzseek -----
--    Sets the starting position for the next gzread or gzwrite on the
-- given compressed file. The offset represents a number of bytes in the
-- uncompressed data stream. The whence parameter is defined as in lseek(2);
-- the value SEEK_END is not supported.
--   If the file is opened for reading, this function is emulated but can be
-- extremely slow. If the file is opened for writing, only forward seeks are
-- supported; gzseek then compresses a sequence of zeroes up to the new
-- starting position.
--
--    gzseek returns the resulting offset location as measured in bytes from
-- the beginning of the uncompressed stream, or -1 in case of error, in
-- particular if the file is opened for writing and the new starting position
-- would be before the current position.
global function gzseek(gzFile file, integer offset, integer whence)
	return c_func(Gzseek, {file, offset, whence})
end function

----- gzrewind -----
--   Rewinds the given file. This function is supported only for reading.
-- gzrewind(file) is equivalent to gzseek(file, 0, SEEK_SET)
global function gzrewind(gzFile file)
	return c_func(Gzrewind, {file})
end function

----- gztell -----
--   Returns the starting position for the next gzread or gzwrite on the
-- given compressed file. This position represents a number of bytes in the
-- uncompressed data stream.
-- gztell(file) is equivalent to gzseek(file, 0, SEEK_CUR)
global function gztell(gzFile file)
	return c_func(Gztell, {file})
end function

----- gzeof -----
--   Returns 1 when EOF has previously been detected reading the given
-- input stream, otherwise zero.
global function gzeof(gzFile file)
	return c_func(Gzeof, {file})
end function

----- gzclose -----
--   Flushes all pending output if necessary, closes the compressed file
-- and deallocates all the (de)compression state. The return value is the zlib
-- error number (see function gzerror below).
global function gzclose(gzFile file)
	return c_func(Gzclose, {file})
end function

----- gzerror -----
--   Returns the error message for the last error which occurred on the
-- given compressed file as {errnum, errormsg} errnum is set to zlib
-- error number. If an error occurred in the file system and not in
-- the compression library, errnum is set to Z_ERRNO and the
-- application may consult errno to get the exact error code.
global function gzerror(gzFile file)
atom lperr, ret
integer errcode, len
sequence errstr
lperr = alloc8(4)
ret = c_func(Gzerror, {file, lperr})
len = 0
while peek(ret + len) do
	len += 1
end while
errstr = peek({ret, len})
errcode = peek4u(lperr)
free(lperr)
	return {errcode, errstr}
end function

-------------------------- Checksum Functions ---------------------------------
--                                                                           --
--       These functions are not related to compression but are exported     --
--     anyway because they might be useful in applications using the         --
--     compression library.                                                  --
-------------------------------------------------------------------------------

----- adler32 -----
--   Update a running Adler-32 checksum with the bytes in buf and
-- return the updated checksum. If buf is {}, this function returns
-- the required initial value for the checksum.
-- An Adler-32 checksum is almost as reliable as a CRC32 but can be computed
-- much faster. Usage example:
--
--   adler = adler32(0, {})
--
--   buffer = {}
--   while not(equal(buffer, -1)) do
--     buffer = read_buffer(len)
--     adler = adler32(adler, buffer)
--   end while
--   if adler != original_adler then
--     error()
--   end if
global function adler32(atom adler, sequence buf)
integer len
atom ret, lpbuf
len = length(buf)
if equal(buf, {}) then
	buf = Z_NULL
else
	lpbuf = alloc8(len)
	poke(lpbuf, buf)
end if
ret = c_func(Adler32, {adler, lpbuf, len})
if not(equal(buf, {})) then
	free(lpbuf)
end if
	return ret
end function

----- crc32 -----
--   Update a running crc with the bytes in buf and return the updated
-- crc. If buf is {}, this function returns the required initial value
-- for the crc. Pre- and post-conditioning (one's complement) is performed
-- within this function so it shouldn't be done by the application.
-- Usage example:
--
--   crc = crc32(0, {})
--
--   buffer = {}
--   while not(equal(buffer, -1)) do
--     buffer = read_buffer(len)
--     crc = crc32(crc, buffer)
--   end while
--   if crc != original_crc then
--     error()
--   end if
global function crc32(atom crc, sequence buf)
integer len
atom ret, lpbuf
len = length(buf)
if equal(buf, {}) then
	buf = Z_NULL
else
	lpbuf = alloc8(len)
	poke(lpbuf, buf)
end if
ret = c_func(Crc32, {crc, lpbuf, len})
if not(equal(buf, {})) then
	free(lpbuf)
end if
	return ret
end function

-------------------------- Various Hacks, Don't Look :) -----------------------

global function zError(integer err)
integer len
atom ret
ret = c_func(ZError, {err})
len = 0
while peek(ret + len) do
	len += 1
end while
	return peek({ret, len})
end function

global function inflateSyncPoint(z_streamp z)
	return c_func(InflateSyncPoint, {z})
end function

-- I dunno what this is....
global function get_crc_table()
	return c_func(Get_crc_table, {})
end function
